"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const Errors_1 = require("../Errors/Errors");
/**
 * Metadata key to mark class to be stored as redis hash
 */
exports.REDIS_ENTITY = Symbol("Redis Hash");
/**
 * Metadata key to store redis properties metadatas
 */
exports.REDIS_PROPERTIES = Symbol("Redis properties");
/**
 * Metadata key to keep serialized initial value of property from redis. For collections this will be link to the corresponding hashmap/set
 */
exports.REDIS_VALUE = Symbol("Redis initial value");
/**
 * Metadata key to keep the initial state of redis collections (map/set)
 */
exports.REDIS_COLLECTION_VALUE = Symbol("Redis initial collection value");
/**
 * Return true if given object is redis entity like
 *
 * @export
 * @param entity
 * @returns
 */
function isRedisEntity(entity) {
    if (typeof entity !== "object" || entity === null) {
        return false;
    }
    const metadata = Reflect.getMetadata(exports.REDIS_ENTITY, entity.constructor);
    return !!metadata;
}
exports.isRedisEntity = isRedisEntity;
/**
 * Return entity name
 *
 * @export
 * @param entityOrEntityClass
 * @returns
 */
function getEntityName(entityOrEntityClass) {
    const metadata = Reflect.getMetadata(exports.REDIS_ENTITY, typeof entityOrEntityClass === "function" ? entityOrEntityClass : entityOrEntityClass.constructor);
    return metadata;
}
exports.getEntityName = getEntityName;
/**
 * Return entity id
 *
 * @export
 * @param entity
 * @returns
 */
function getEntityId(entity) {
    if (!isRedisEntity(entity)) {
        return undefined;
    }
    const metadatas = Reflect.getMetadata(exports.REDIS_PROPERTIES, entity.constructor);
    const idMetadata = metadatas && metadatas.find(val => !val.isRelation && val.isIdentifyColumn);
    if (!idMetadata) {
        return undefined;
    }
    if (typeof entity[idMetadata.propertyName] === "number" || typeof entity[idMetadata.propertyName] === "string") {
        return entity[idMetadata.propertyName];
    }
    return undefined;
}
exports.getEntityId = getEntityId;
/**
 * Returns redis properties metadata
 *
 * @export
 * @param entityOrEntityClass
 * @returns
 */
function getEntityProperties(entityOrEntityClass) {
    return Reflect.getMetadata(exports.REDIS_PROPERTIES, typeof entityOrEntityClass === "function" ? entityOrEntityClass : entityOrEntityClass.constructor);
}
exports.getEntityProperties = getEntityProperties;
/**
 * Return full hash id
 *
 * @export
 * @param entityOrEntityClass
 * @param [hashId]
 * @returns
 */
function getEntityFullId(entityOrEntityClass, hashId) {
    const name = getEntityName(entityOrEntityClass);
    const id = typeof hashId !== "undefined"
        ? hashId
        : typeof entityOrEntityClass === "object"
            ? getEntityId(entityOrEntityClass)
            : undefined;
    if (name && typeof id !== "undefined") {
        return `e:${name}:${id}`;
    }
}
exports.getEntityFullId = getEntityFullId;
/**
 * Return relation type from relation type func from metadata
 *
 * @export
 * @param entityOrEntityClass
 * @param metadata
 * @returns
 */
function getRelationType(entityOrEntityClass, metadata) {
    const entityName = typeof entityOrEntityClass === "function" ? entityOrEntityClass : entityOrEntityClass.constructor;
    if (!metadata.isRelation) {
        throw new Errors_1.MetadataError(entityName, `${metadata.propertyName} is not a relation`);
    }
    const typeOrTypes = metadata.relationTypeFunc();
    const type = Array.isArray(typeOrTypes) ? typeOrTypes[0] : typeOrTypes;
    if (typeof type !== "function") {
        throw new Errors_1.MetadataError(entityName, `Invalid relation type ${type} for property ${metadata.propertyName}`);
    }
    if (typeof Reflect.getMetadata(exports.REDIS_ENTITY, type) === "undefined") {
        throw new Errors_1.MetadataError(entityName, `Relation entity ${type.name} must be decorated with @Entity`);
    }
    return type;
}
exports.getRelationType = getRelationType;
//# sourceMappingURL=Metadata.js.map